/*************************************************************************/
/*                                                                       */
/*                            LD-CELP  G.728                             */
/*                                                                       */
/*    Low-Delay Code Excitation Linear Prediction speech compression.    */
/*                                                                       */
/*    Code edited by Michael Concannon.                                  */
/*    Based on code written by Alex Zatsman, Analog Devices 1993         */
/*                                                                       */
/*************************************************************************/


/********************************* Perceptual Weighting Filter **/

#include "common.h"
#include "parm.h"
#include "fast.h"
#include "prototyp.h"


/************ Prototypes ************/

void sound_g728_pwfilter2(sound_g728_t *this, real QMEM input[], real output[]);
void sound_g728_sf_zresp(sound_g728_t *this, real output[]);
void sound_g728_pwf_zresp(sound_g728_t *this, real input[], real output[]);
void sound_g728_zresp(sound_g728_t *this, real output[]);
void sound_g728_mem_update (sound_g728_t *this, real input[], real output[]);
static real log_rms(real input[]);
real sound_g728_predict_gain(sound_g728_t *this);
void sound_g728_update_gain(sound_g728_t *this, real input[], real *lgp);
void sound_g728_init_gain_buf(sound_g728_t *this);

/************************************/

/* Filter Memory in reverse order, i.e. firmem[0] is the most recent */

void sound_g728_pwfilter2(sound_g728_t *this, real QMEM input[], real output[])
{
	int k;
	real out;
	
	RSHIFT(this->firmem, LPCW, IDIM);
	for(k=0; k<IDIM; k++)
		this->firmem[k] = input[IDIM-1-k];
	RSHIFT(this->iirmem, LPCW, IDIM);

	for (k=0; k<IDIM; k++) {
		out = this->firmem[IDIM-1-k];		/* pwf_z_coeff[0] is always 1.0 */
		out += DOTPROD(this->firmem+IDIM-k, this->pwf_z_coeff+1, LPCW);
		out -= DOTPROD(this->iirmem+IDIM-k, this->pwf_p_coeff+1, LPCW);
		this->iirmem[IDIM-1-k] = out;
		output[k] = out;
	}
}

/*  Synthesis and Perceptual Weighting Filter. */

/** Updateable coefficients **/

void sound_g728_sf_zresp(sound_g728_t *this, real output[])
{
	int k,j;

#ifdef PIPELINE
	for(j=LPC-1; j>=0; j--)
		this->statelpc[j+IDIM] = this->statelpc[j];
	
	{
		real STATE_MEM * sjpk = this->statelpc + LPC+IDIM-1; 
		real out = 0.0;
		for (k=0; k<IDIM; k++) {
			real COEFF_MEM *ajp = this->sf_coeff+LPC;
			real sj, aj;
			real STATE_MEM *sjp;
			out = 0.0;
			sjp = sjpk;
			sj = *sjp--;
			aj = *ajp--;
			for (j=LPC-2; j>=0; j--) {
				out -= sj*aj;
				sj = *sjp--;
				aj = *ajp--;
			}
			output[k] = out - sj*aj;
			*sjp--= output[k];
			sjpk--;
		}
	}
#else
/** This is un-pipelined version of the above. Kept for reference **/
	for(j=LPC-1; j>=0; j--)
		this->statelpc[j+IDIM] = this->statelpc[j];
	for (k=0; k<IDIM; k++) {
		real out = 0.0, sj, aj;
		sj = this->statelpc[LPC+IDIM-k-1];
		aj = this->sf_coeff[LPC];
		for (j=LPC-2; j>=1; j--) {
			out -= sj*aj;
			sj = this->statelpc[IDIM-k+j];
			aj = this->sf_coeff[j+1];
		}
		output[k] = out - sj*aj-this->statelpc[IDIM-k] * this->sf_coeff[1];
		this->statelpc[IDIM-1-k] = output[k];
	}
	return;
#endif
}

void sound_g728_pwf_zresp(sound_g728_t *this, real input[], real output[])
{
	int j,k;
	real tmp;

#ifdef PIPELINE
	for (k=0; k<IDIM; k++) {
		tmp = input[k];
		for (j=LPCW-1; j>=1; j--) {
			input[k] += this->zirwfir[j] * this->pwf_z_coeff[j+1];
			this->zirwfir[j] = this->zirwfir[j-1];
		}
		input[k] += this->zirwfir[0] * this->pwf_z_coeff[1];
		this->zirwfir[0] = tmp;
		for (j=LPCW-1; j>=1; j--) {
			input[k] -= this->zirwiir[j] * this->pwf_p_coeff[j+1];
			this->zirwiir[j] = this->zirwiir[j-1];
		}
		output[k] = input[k] - this->zirwiir[0] * this->pwf_p_coeff[1];
		this->zirwiir[0] = output[k];
	}
#else
   /** Un-pipelined version, kept for reference **/
	for (k=0; k<IDIM; k++) {
		tmp = input[k];
		for (j=LPCW-1; j>=1; j--) {
			input[k] += this->zirwfir[j] * this->pwf_z_coeff[j+1];
			this->zirwfir[j] = this->zirwfir[j-1];
		}
		input[k] += this->zirwfir[0] * this->pwf_z_coeff[1];
		this->zirwfir[0] = tmp;
		for (j=LPCW-1; j>=1; j--) {
			input[k] -= this->zirwiir[j] * this->pwf_p_coeff[j+1];
			this->zirwiir[j] = this->zirwiir[j-1];
		}
		output[k] = input[k] - this->zirwiir[0] * this->pwf_p_coeff[1];
		this->zirwiir[0] = output[k];
	}
#endif   
}


void sound_g728_zresp(sound_g728_t *this, real output[])
{
	real temp[IDIM];
	sound_g728_sf_zresp(this, temp);
	sound_g728_pwf_zresp(this, temp, output);
}


void sound_g728_mem_update (sound_g728_t *this, real input[], real output[])
{
	int i,k;
	real temp[IDIM], a0, a1, a2;
	real STATE_MEM *t2 = this->zirwfir;
	t2[0] = temp[0] = input[0];
	for (k=1; k<IDIM; k++) {
		a0 = input[k];
		a1 = a2 = 0.0;
		for (i=k; i>= 1; i--) {
			t2[i] = t2[i-1];
			temp[i] = temp[i-1];
			a0 -=   this->sf_coeff[i] * t2[i];
			a1 += this->pwf_z_coeff[i] * t2[i];
			a2 -= this->pwf_p_coeff[i] * temp[i];
		}
		t2[0] = a0;
		temp[0] = a0+a1+a2;
	}
	
	for (k=0; k<IDIM; k++) {
		this->statelpc[k] += t2[k];
		if (this->statelpc[k] > Max) this->statelpc[k] = Max;
		else if (this->statelpc[k] < Min) this->statelpc[k] = Min;
		this->zirwiir[k] += temp[k];
	}
	for (i=0; i<LPCW; i++) this->zirwfir[i] = this->statelpc[i];
	for (k=0; k<IDIM; k++) output[k] = this->statelpc[IDIM-1-k];
	return;
}

#include <math.h>

# define LOG10(X) log10(X)
# define EXP10(X) pow(10,X)

/*********************************************** The Gain Predictor */

static real log_rms(real input[])
{
	real etrms=0.0;
	int k;
	for(k=0; k<IDIM; k++) etrms += input[k]*input[k];
	etrms /= IDIM;
	if (etrms<1.0) etrms = 1.0;
	etrms = 10.0*log10(etrms);
	return (etrms);
}

real sound_g728_predict_gain(sound_g728_t *this)
{
	int i;
	real new_gain = GOFF;
	real temp;
	
	for (i=1;i<=LPCLG;i++) {
		temp = this->gp_coeff[i] * this->gain_input[LPCLG-i];
		new_gain -= temp;
	}
	if (new_gain <  0.0) new_gain = 0.0;
	if (new_gain > 60.0) new_gain = 60.0;
	new_gain = EXP10(0.05*new_gain);
	return (new_gain);
}

void sound_g728_update_gain(sound_g728_t *this, real input[], real *lgp)
{
	int i;

	*lgp = log_rms(input) - GOFF;
	for (i=0; i<LPCLG-1; i++) this->gain_input[i] = this->gain_input[i+1];
	this->gain_input[LPCLG-1] = *lgp;
}

void sound_g728_init_gain_buf(sound_g728_t *this)
{
	int i;
	
	for(i=0;i<LPCLG;i++) this->gain_input[i] = -GOFF;
	for(i=0;i<QSIZE/IDIM;i++) this->log_gains[i] = -GOFF;
}
